import numpy as np
import scipy.stats
import itertools
import pandas as pd
from six import iteritems


class AnalysisTools:
    """A container for methods involving the creation of summary data."""    
    
    def __init__(self, model):
        """Creates a class containing tools for analyzing model outputs.
        =========
        Arguments
        ---------
        model: the ExemplarModel object in which the output was obtained.
        """
        
        #Backpointer
        self.model = model
        
        #Filters based on word frequency
        self.identFilter = lambda word: True
        self.hfFilter = lambda word: word.frequency in model.freqBins["HF"]
        self.mhfFilter = lambda word: word.frequency in model.freqBins["MHF"]
        self.mlfFilter = lambda word: word.frequency in model.freqBins["MLF"]
        self.lfFilter = lambda word: word.frequency in model.freqBins["LF"]
        
        
    def getExemplarValues(self, exemplars):
        """Return an array containing the values of a list of exemplars.
        =========
        Arguments
        ---------
        exemplars: a list of Exemplar objects.
        =========
        Return: a np.array of exemplar values.
        """
        
        values = np.array([exemplar.value for exemplar in exemplars])
        
        return values
        
        
    def getOverlapStatistics(self, cat1, cat2, filterFunc = None):
        """Get the overlap width, number and area between two categories.
        Overlap width is represented as the span of the overlapping region.
        Overlap number is represented as a proportion of the total number of 
        exemplars in the two categories. Area is represented as the area strictly
        between the two exemplar histograms, as a proportion of the total area
        spanned by the two histograms.
        =========
        Arguments
        ---------
        cat1: the CategoryCloud object representing the first category.
        cat2: the CategoryCloud object representing the second category.
        filterFunc: a function mapping from a WordType object to a Boolean
                    indicating whether that word is included (typically based
                    on word frequency).
        =========
        Return: a triple specifying overlap width, overlap number, and overlap area.
        """
        
        #The defauly filter is an identity filter, which includes all words
        if filterFunc is None:
            filterFunc = self.identFilter
        
        #Get the total number of exemplars in both categories (before filtering)
        Nexemplars = cat1.totalWordFrequency + cat2.totalWordFrequency 
        
        #Get filtered exemplars from each category
        cat1exemplars = self.getFilteredWordCloud(cat1, filterFunc)
        cat2exemplars = self.getFilteredWordCloud(cat2, filterFunc)
        
        #Convert the exemplars to density histograms
        cat1density = self.getExemplarHistogram(cat1exemplars)
        cat2density = self.getExemplarHistogram(cat2exemplars)
        
        #Get the range of acoustic values spanned by each category,
        #and the extents of that span
        cat1values = self.model.spaceValues[cat1density > 0]
        cat2values = self.model.spaceValues[cat2density > 0]
        cat1min = cat1values.min()
        cat1max = cat1values.max()
        cat2min = cat2values.min()
        cat2max = cat2values.max()
        
        #Calculate the overlap width based on category spans
        minValue = max(cat1min, cat2min)
        maxValue = min(cat1max, cat2max)
        overlapWidth = maxValue - minValue
        
        #Get the exemplars within the overlapping region
        overlapRange = np.logical_and(self.model.spaceValues <= maxValue, self.model.spaceValues >= minValue)
        cat1overlap = cat1density[overlapRange]
        cat2overlap = cat2density[overlapRange]
        
        #Calculate the number of exemplars in the overlapping region, as a proportion
        overlapNum = cat1overlap.sum() + cat2overlap.sum()
        overlapNumProp = overlapNum / Nexemplars
        
        #Calculate the area under both distributions in the overlapping region, as a proportion
        overlapArea = np.minimum(cat1overlap, cat2overlap).sum()
        overlapAreaProp = overlapArea / (Nexemplars - overlapArea)                    
                        
        return overlapWidth, overlapNumProp, overlapAreaProp


    def getExemplarHistogram(self, exemplars):
        """Convert a list of exemplars to an array representing a histogram of their values.
        =========
        Arguments
        ---------
        exemplars: a list of Exemplar objects.
        =========
        Return: a np.array representing a histogram over the exemplar space.
        """
        
        #Initialize the distribution
        hist = np.zeros(self.model.spacePoints)
        
        #Count the exemplars at each point
        for exemplar in exemplars:
            hist[self.model.spaceIndices[exemplar.value]] += 1
            
        return hist
        
        
    def getFilteredExemplarHistogram(self, category, filterFunc):
        """Get a histogram of a filtered category.
        =========
        Arguments
        ---------
        category: a CategoryCloud object.
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.
        =========
        Return: a np.array representing a histogram over the exemplar space.
        """
        
        #Get a list of the filtered exemplars
        exemplars = self.getFilteredWordCloud(category, filterFunc)
        
        #Convert the list to a histogram
        hist = self.getExemplarHistogram(exemplars)
            
        return hist        
        
        
    def getFilteredExemplarHistograms(self, agent, filterFunc):
        """Returns a dict mapping from each category to an exemplar histogram,
        taken across words that satisfy a boolean filter function.
        =========
        Arguments
        ---------
        agent: the Agent object in which the categories are situated.
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.
        =========
        Return: a dictionary mapping from category names to np.arrays representing
                histograms over the acoustic space.
        """
        
        #Initialize a dictionary of histograms
        histDict = dict()
        
        #Get the histogram for each category and add it to the dictionary
        for category in agent.categories:
            histDict[category.name] = self.getFilteredExemplarHistogram(category, filterFunc)
            
        return histDict        
        
        
    def getActivationField(self, values, category):
        """Convert a list of exemplar values to an array representing their activation field.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object in which the exemplars reside.
        =========
        Return: an np.array representing the activation field.
        """
        
        #Initialize the activation field
        field = np.zeros(self.model.spacePoints)
        
        #Add the activation windows centered at every exemplar value
        for value in values:
            field += category.agent.perWindows[value]
            
        return field
        
        
    def getFilteredActivationField(self, category, filterFunc):
        """Get the activation field of a filtered category.
        =========
        Arguments
        ---------
        category: the CategoryCloud object to get the activation field of.
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.
        =========
        Return: an np.array representing the activation field.        
        """ 
        
        #For the unfiltered category, read the activation field directly
        if filterFunc is self.identFilter:
            field = category.actSpace
      
        #If filtering, apply the filter, and then get the activation field of the filtered exemplars  
        else:
            exemplars = self.getFilteredWordCloud(category, filterFunc)
            values = self.getExemplarValues(exemplars)
            field = self.getActivationField(values, category)
            
        return field  
                
        
    def getFilteredActivationFields(self, agent, filterFunc):
        """Returns a dict mapping from each category to an activation field,
        taken across words that satisfy a boolean filter function.
        =========
        Arguments
        ---------
        agent: the Agent object to get the activation fields for.
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.
        =========
        Return: a dictionary mapping from category names to np.arrays representing
                the activation fields.         
        """
        
        #Intialize the dictionary
        fieldDict = dict()
        
        #Add the filtered activation field for each category to the dictionary
        for category in agent.categories:
            fieldDict[category.name] = self.getFilteredActivationField(category, filterFunc)
            
        return fieldDict        
         
        
    def getFilteredWordCloud(self, category, filterFunc):
        """Return a list of exemplars whose words pass some boolean-function test.
        =========
        Arguments
        ---------
        category: the CategoryCloud object to get the exemplars from.
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.
        =========
        Return: a list of exemplars from the category whose words pass the filter.           
        """
        
        #Initialize the list
        exemplars = list()
        
        #Add the exemplars for each word which passes the filter
        for wordType in category.wordTypes:
            if filterFunc(wordType):
                exemplars += wordType.exemplars
                
        return exemplars
        
              
    def getWidth(self, values, category):
        """Get the width of an array of exemplar values, measured as estimated standard deviation.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the standard deviation of the exemplar values.
        """
        
        width = values.std()
        
        return width      
    
    
    def getCentroid(self, values, category):
        """Get the centroid of an array of exemplar values, measured as mean exemplar value.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the mean of the exemplar values.        
        """

        centroid = values.mean()
        
        return centroid
        
    
    def getModeAndPeak(self, values, category):
        """Get the mode and peak of an array of exemplar values, measured as location and size of activation peak.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: a pair containing the point in the exemplar space with greatest
                activation of the category, and the activation at that point.
        """
        
        #Get the activation field for the category
        activation = self.getActivationField(values, category)
        
        #Get the mode and peak of the activation field
        mode = self.model.spaceValues[activation.argmax()]
        peak = activation.max()
        
        return mode, peak   
        

    def getMinValue(self, values, category):
        """Get the minimum value among an array of exemplar values.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the minimum of the exemplar values.        
        """
        
        minValue = values.min()
        
        return minValue
        

    def getMaxValue(self, values, category):
        """Get the maximum value among an array of exemplar values.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the maximum of the exemplar values.        
        """
        
        maxValue = values.max()
        
        return maxValue  
        
              
    def getSkewness(self, values, category):
        """Get the skewness of a distribution of exemplar values.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the skewness of the exemplar values.        
        """     
        
        skewness = scipy.stats.skew(values)
        
        return skewness 
        
        
    def getKurtosis(self, values, category):
        """Get the kurtosis of a distribution of exemplar values.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the kurtosis of the exemplar values.        
        """     
        
        kurtosis = scipy.stats.kurtosis(values)
        
        return kurtosis 
        
        
    def getMedian(self, values, category):
        """Get the median of a distribution of exemplar values.
        =========
        Arguments
        ---------
        values: a list of exemplar values.
        category: the CategoryCloud object from which the exemplar values are taken.
        =========
        Return: the median of the exemplar values.        
        """     
        
        median = np.median(values)
        
        return median        
        
        
    def getCategoryStatistics(self, category, statFuncDict, filterFunc = None):
        """Get a dictionary of summary statistic values of a filtered category.
        =========
        Arguments
        ---------
        category: the CategoryCloud object to get summary statistics for.
        statFuncDict: a dictionary mapping from string names (or tuples of string
                      names) for return values of statistical summary functions 
                      to those statistical summary functions. Each function takes
                      a list of exemplar values and a CategoryCloud object as input 
                      and returns a statistic (or tuple of statistics).
        filterFunc: a filter function, mapping from WordTypes to a Boolean indicating
                    whether or not they pass the filter.        
        =========
        Return: a dictionary mapping from string names of statistics to their values.
                Tuples are unpacked in this dictionary.
        """

        #The default filter is the identity filter, returning all words in the category
        if filterFunc is None:
            filterFunc = self.identFilter
            
        #Get the filtered exemplars and their values
        exemplars = self.getFilteredWordCloud(category, filterFunc)
        values = self.getExemplarValues(exemplars)            
            
        #Initialize the summary statistic dictionary
        statsDict = dict()
        
        #Iterate through the summary statistic functions
        for statFuncName, statFunc in iteritems(statFuncDict):

            #If there is no filtering, retrieve the mode and peak efficiently
            if statFunc is self.getModeAndPeak and filterFunc is self.identFilter:
                activation = category.actSpace
                mode = self.model.spaceValues[activation.argmax()]
                peak = activation.max()
                stat = mode, peak  
            
            #Apply each function to the category
            else:   
                stat = statFunc(values, category)
            
            #Unpack tuple statistics into single entries in the dictionary
            if isinstance(statFuncName, tuple):
                statsDict.update(dict(zip(statFuncName, stat)))
            
            #Add the statistic to the dictionary
            else:
                statsDict[statFuncName] = stat
        
        return statsDict
        
    
    def makeSummaryDict(self):
        """Create a dictionary in which the model run summary will be stored.
        =========
        Return: a dictionary mapping from agent names to dictionaries of summary
                statistics at various different iterations.        
        """
        
        #Initialize the summary dictionary
        summaryDict = dict()
        
        #Iterate through the Agent objects in the model
        for agent in self.model.agents:
            
            #Add the agent to the dictionary
            summaryDict[agent.name] = dict()
            summaryDict[agent.name]["iteration"] = list()
        
            #For each category within the agent, add summary statistics for 
            #centroid, width, mode, peak, skew, and kurtosis,
            #for the whole category, the high-frequency sub-distribution, and the low-frequency sub-distribution
            for categoryName in agent.categoryStore:
                for freqName in ["",".hf",".lf"]:
                    for statName in [".centroid", ".width", ".mode", ".peak", ".skew", ".kurtosis"]:
                        summaryDict[agent.name]["%s%s%s" % (categoryName, freqName, statName)] = list()
 
            #For each pair of categories within the agent, add summary statistics for 
            #overlap between the categories: width, number, and area                              
            for indexPair in itertools.combinations(range(agent.numCategories), 2):
                (cat1, cat2) = sorted([agent.categories[indexPair[0]], agent.categories[indexPair[1]]], key= lambda c: c.name)                         
                for overlapStat in ['overlapWidth', 'overlapNum', 'overlapArea']:
                    summaryDict[agent.name]['%s%s.%s' % (cat1.name, cat2.name, overlapStat)] = list()
            
        return summaryDict   
        
                    
    def updateSummaryDict(self, summaryDict, iterNum):
        """Update the supplied summary dict with the current statistics.
        =========
        Arguments
        ---------
        summaryDict: the existing summary dictionary, mapping from agent names to 
                     dictionaries of summary statistics at previous iterations.
        iterNum: the current iteration number.
        =========
        Return: an updated dictionary mapping from agent names to dictionaries of 
                summary statistics at previous and current iterations.       
        """
        
        #The three word filters used: all words, high-frequency only, low-frequency only
        freqFuncDict = {"":None, ".hf":self.hfFilter, ".lf":self.lfFilter}

        #The single category summary statistics: centroid, width, mode, peak, skew, kurtosis
        statFuncDict = {".centroid": self.getCentroid, ".width": self.getWidth, (".mode", ".peak"): self.getModeAndPeak, ".skew": self.getSkewness, ".kurtosis": self.getKurtosis}
        
        #Iterate through the agents
        for agent in self.model.agents:
            
            #Add the current iteration number
            summaryDict[agent.name]["iteration"].append(iterNum)
            
            #For each category, add the current value of each summary statistic for each filter
            for categoryName in agent.categoryStore:
                for freqName in freqFuncDict:
                    freqFunc = freqFuncDict[freqName]
                    statsDict = self.getCategoryStatistics(agent.getCategory(categoryName), statFuncDict, freqFunc)
                    for statName, stat in iteritems(statsDict):
                        summaryDict[agent.name]["%s%s%s" % (categoryName, freqName, statName)].append(stat)
                       
            #For each pair of categories, add the current value of the three overlap statistics for each filter     
            for indexPair in itertools.combinations(range(agent.numCategories), 2):
                (cat1, cat2) = sorted([agent.categories[indexPair[0]], agent.categories[indexPair[1]]], key= lambda c: c.name)
                overlapWidth, overlapNum, overlapArea = self.getOverlapStatistics(cat1, cat2)
                summaryDict[agent.name]['%s%s.overlapWidth' % (cat1.name, cat2.name)].append(overlapWidth)
                summaryDict[agent.name]['%s%s.overlapNum' % (cat1.name, cat2.name)].append(overlapNum)
                summaryDict[agent.name]['%s%s.overlapArea' % (cat1.name, cat2.name)].append(overlapArea)
                        
        return summaryDict 
        
    
    def makeSummaryDataFrame(self, summaryDict):
        """Convert the provided summary dictionary to a summary dataframe.
        =========
        Arguments
        ---------
        summaryDict: the summary dictionary, mapping from agent names to 
                     dictionaries of summary statistics at various iterations.
        =========
        Return: a pandas DataFrame object containing the summary statistics for
                all agents at all iterations.        
        """
        
        #If there is a single agent, convert that agent's summary dictionary
        if self.model.numAgents == 1:
            agent = self.model.agents[0]
            summaryDataFrame = pd.DataFrame(summaryDict[agent.name])
        
        #If there are multiple agents, collapse them all into a single data frame
        else:
            summaryDataFrame = pd.Panel(summaryDict).transpose(2,0,1).to_frame().reset_index()
            summaryDataFrame = summaryDataFrame.drop("minor", 1)
            summaryDataFrame = summaryDataFrame.rename(columns={"major":"agent"})
        
        return summaryDataFrame