import json
import os
import re
import itertools
import exemplar_representations as representations
from builtins import range
from six import iteritems


class IOTools:
    """A container for methods involving input and output files."""    
    
    def __init__(self, model):
        """Create a class containing tools for interfacing with files.
        =========
        Arguments
        ---------
        model: the ExemplarModel object for taking input and providing output.
        """
        
        self.model = model
        

    def loadData(self, agent, dataFilename="initial_data.json", makeAdjustments=True, shuffleWords=False, clearExemplars=True, shuffleFrequencies=False):
        """Load category, word and exemplar data from a single file,
        and use it to initialize a given agent with exemplars.
        =========
        Arguments
        ---------
        agent: the Agent object into which exemplars will be fed.
        dataFilename: the name of the intial data json file.
        makeAdjustments: a Boolean flag indicating whether to adjust category
                         width and separating using the parameters sigma and mu.
        shuffleWords: a Boolean flag indicating whether to shuffle exemplars across
                      word types within a frequency class on different runs.
        clearExemplars: a Boolean flag indicating whether to clear any existing
                        exemplars in the agent before adding new ones.
        shuffleFrequencies: a Boolean flag indicating whether to shuffle the 
                            frequencies of word types within the same category
                            on different runs.
        """
        
        #Get parameters for adjusting category width and separation
        scaling = 1
        categoryAdjustSize = 0
        if makeAdjustments:
            try:
                scaling = agent.initCategoryWidthSD
            except:
                pass
            try:
                categoryAdjustSize = agent.initCategoryAdjustSize
            except:
                pass 
        
        #Set the initialization data filename as an attribute, in case of reloading later
        self.initDataFilename = dataFilename        
        
        #Read data from file
        try:
            with open(dataFilename, mode='r', encoding='utf-8') as inFile:
                data = json.load(inFile)
        except:
            try:
                with open(dataFilename, mode='rb') as inFile:
                    data = json.load(inFile)
            except:
                raise Exception("Cannot load file %s" % dataFilename)
            
        #Clear pre-existing exemplars if required
        if clearExemplars:
            for category in agent.categories:
                category.clearExemplars()      
        
        #Parse each category in the data
        for categoryName in data:
            categoryData = data[categoryName]
            
            #Calculate category-specific adjustment
            categoryAdjust = categoryAdjustSize * categoryData['adjustment']
            
            #Add or retrieve CategoryCloud object for the category
            if categoryName not in agent.categoryStore:
                category = representations.CategoryCloud(categoryName, categoryData['bias'], agent)
                agent.addCategory(category)
            else:
                category = agent.categoryStore[categoryName]
            
            #Parse each word in the category data
            for wordFrame in categoryData['words']:
                wordData = categoryData['words'][wordFrame]
                
                #Add or retrieve WordType obejt for the word
                if wordFrame not in category.wordTypeStore:
                    wordFreq = wordData['frequency']
                    word = representations.WordType(wordFrame, wordFreq, category)
                    category.addWordType(word)
                    
                    #Update max frequency counter for the agent
                    if wordFreq > agent.maxFreq:
                        agent.maxFreq = wordFreq
                else:
                    word = category.getWordType(wordFrame)
                    
                    # Reset frequency in case it has been disrupted by shuffling
                    if word.frequency != wordData['frequency']:
                        word.frequency = wordData['frequency']
                
                #Parse each exemplar in the word data
                for exemplarValue in wordData['exemplars']:
                        
                    #Adjust for category width and separation if required
                    if scaling != 1:
                        exemplarValue = exemplarValue * scaling      
                    if categoryAdjust != 0:
                        exemplarValue += categoryAdjust
                    
                    #Create Exemplar object
                    exemplar = representations.Exemplar(exemplarValue, wordFrame, None)
                    exemplar.model = self.model
                    
                    #Discretize the exemplar value and bring it into the space
                    if scaling != 1 or categoryAdjust != 0:
                        exemplar.discretize()
                        exemplar.bringIntoSpace()
                    
                    #Add the exemplar to the word
                    word.addExemplar(exemplar)                        
            
            #Shuffle frequencies of word types within the same category if required        
            if shuffleFrequencies:
                category.shuffleFrequencies()
            
            #Shuffle exemplars across word types within the same frequency class if required
            if shuffleWords:
                category.shuffleWords()                               
        
             
    def loadMultiple(self, agent, dataFilenames=None, dataPath="Dumps", fileNameSuffix="iter0"):
        """Load category, word and exemplar data from multiple files,
        and use it to initialize a given agent with exemplars.
        No adjustments to category width or separation are possible, there is no
        shuffling of exemplars or frequencies across word types, and pre-existing
        exemplars in the agent are cleared.
        =========
        Arguments
        ---------
        agent: the Agent object into which exemplars will be fed.
        dataFilenames: a list containing the names of the intial data json files,
                       or a string representing a single filename, or None (in
                       which case all json files in the directory with a given
                       suffix are used).
        dataPath: a string path to the directory where initialization files are stored.
        filenameSuffix: a string representing the common suffix (minus extension) 
                        of the filenames to be loaded. Can use regular expressions.
        """
        
        #If there is a single filename provided, just use that filename
        if isinstance(dataFilenames, str):
            dataFilenames = [dataFilenames]
        #If the filename is None, use all filenames in the directory matching the suffix pattern
        elif dataFilenames is None:
            dataFilenames = ["%s/%s" % (dataPath, f) for f in os.listdir(dataPath) if re.search("%s\.json" % fileNameSuffix, f)]      
        
        #Clear pre-existing exemplars
        for category in agent.categories:
            category.clearExemplars()
        
        #Load data from each file
        for dataFilename in dataFilenames:
            self.loadData(agent, dataFilename=dataFilename, makeAdjustments=False, shuffleWords=False, clearExemplars=False, shuffleFrequencies=False)
                    
        #Set the initialization data filename attribute to be None, since no single filename was used
        self.initDataFilename = None
            
                   
    def dumpData(self, agent, filename):
        """Dump the state of a single agent (categories, words, exemplars) to file
        in json format.
        =========
        Arguments
        ---------
        agent: the Agent object whose data will be dumped.
        filename: a string representing the name of the file to be dumped.
        """
        
        #Set up the state data in the format of initialization data
        #Adjustment is set to 0 to allow reloading of the data later
        data = {
                 category.name: 
                   {
                     "bias": category.bias,
                     "adjustment": 0,
                     "words": 
                       {
                         word.frame: 
                           {
                             "frequency": word.frequency,
                             "exemplars": [exemplar.value for exemplar in word.exemplars]
                           }
                         for word in category.wordTypes
                       }
                   }
                 for category in agent.categories
               }
        
        #Dump to file
        try:
            with open(filename, mode="w", encoding='utf-8') as outFile:       
                json.dump(data, outFile, indent=4)
        except:
            try:
                with open(filename, mode="wb") as outFile:       
                    json.dump(data, outFile, indent=4)
            except:
                raise Exception("Error dumping data to file.")
            
    
    def dumpDataFrame(self, df, path, filename):
        """Dump a pandas DataFrame to file in csv format.
        =========
        Arguments
        ---------
        df: a pandas DataFrame object to be dumped to file.
        path: a string path to the directory where the file will be saved.
        filename: a string representing the filename to be saved (including extension).
        """
        
        df.to_csv("%s/%s" % (path, filename), mode="w", encoding='utf-8', index=False)
        
        

class SettingsParser:
    """A container for methods for parsing and interacting with a settings file.""" 
    
    def __init__(self, filename="settings.json"):
        """Create a class for parsing a settings file.
         =========
        Arguments
        ---------
        filename: a string representing the filename of the json settings file.
        """
        
        #Read the settings file 
        try:
            with open(filename, mode="r", encoding='utf-8') as settingsFile:
                self.settings = json.load(settingsFile)
        except:
            try:
                with open(filename, mode="rb") as settingsFile:
                    self.settings = json.load(settingsFile)
            except:
                raise Exception("Something is wrong with the settings file %s.\nPlease check the name, location, and formatting of the file." % filename)
    
        #Set default values for shuffleWords and shuffleFrequencies flags if not present
        if "shuffleWords" not in self.settings:
            self.settings["shuffleWords"] = (self.settings["runsPerSetting"] > 1)
        if "shuffleFrequencies" not in self.settings:
            self.settings["shuffleFrequencies"] = False
            
        #Set an attribute containing the total number of parameter combinations, including repeated runs    
        self.totalParamCombos = self.settings["repeatsPerRun"]
        for var in self.settings:
            if isinstance(self.settings[var], list):
                self.totalParamCombos *= len(self.settings[var])            
            
            
    def __getitem__(self, varName):
        """Get the setting for a particular variable.
        =========
        Arguments
        ---------
        varName: the string name of the variable to get the setting for.
        """
        
        return self.settings[varName]
        
        
    def paramGenerator(self):
        """Return a generator over parameter values for initializing the model.
        =========
        Returns: a generator over pairs (dumpName, paramValuesDict), where
                 dumpName is a string name for dumps from the present parameter set
                 and paramValuesDict is a dictionary from parameter names to values
                 from the present parameter set.
        """
        
        #Figure out which variables are to be modified in sets
        setVars = [var.split(",") for var in self.settings if "," in var]
            
        #Figure out which variables have multiple values -- excluding those to be modified in sets
        multiVars = [var for var in self.settings if (isinstance(self.settings[var], list) and "," not in var)]
        
        #Set all of the fixed (single) values for initialization
        initVars = {"sigma", "mu", "beta", "iota", "alpha", "delta", "deltaDiff", "tau", "tauDiff", "epsilon", "chi"}
        fixedVars = [var for var in initVars if (var in self.settings and not isinstance(self.settings[var], list))]
        fixedSettings = {var: self.settings[var] for var in fixedVars}
        
        #Repeat the runs the required number of times
        dumpSuffix = ""
        for repeatNum in range(self.settings["repeatsPerRun"]):
            
            #Log the repeat number in the filename suffix for dumps
            if self.settings["repeatsPerRun"] > 1:
                dumpSuffix = "_repeatNum=%s" % repeatNum
        
            #If there are values to be modified in sets,
            #iterate over the combinations of values from different sets
            if len(setVars) > 0:
                for fixedParamSet in itertools.product(*[self.settings[",".join(varSet)] for varSet in setVars]):
                    
                    #Get individual parameter names and values for the current combination
                    varNames = [var for varSet in setVars for var in varSet]
                    paramValues = [param for paramSet in fixedParamSet for param in paramSet]
                
                    #If there are independent parameters to take multiple values, 
                    #iterate over the combinations of values from different parameters
                    if len(multiVars) > 0:
                        for paramSet in itertools.product(*[self.settings[var] for var in multiVars]):
                            yield self.paramComboPair([(varNames, paramValues),(multiVars, paramSet)], fixedSettings, dumpSuffix)
                        
                    #If there are no independent parameters to take multiple values,
                    #just use the parameters modified in sets
                    else:
                        yield self.paramComboPair([(varNames, paramValues)], fixedSettings, dumpSuffix)                     
                        
        
            #If there are independent parameters to take multiple values, 
            #iterate over the combinations of values from different parameters        
            elif len(multiVars) > 0:
                for paramSet in itertools.product(*[self.settings[var] for var in multiVars]):
                    yield self.paramComboPair([(multiVars, paramSet)], fixedSettings, dumpSuffix)
            
            #If there are no parameters to be varied, just use the fixed values
            else:
                yield self.paramComboPair([], fixedSettings, dumpSuffix)
                
                
    def paramComboPair(self, variableParamNamesAndValuesList, fixedParamsDict, dumpSuffix):
        """Return a dump name and parameter value dictionary for a given parameter set.
        =========
        Arguments
        ---------
        variableParamNamesAndValuesList: a list of pairs containing a list of 
                                         parameter names and a list of their current
                                         values for all parameters that take a 
                                         variable value.
        fixedParamsDict: a dictionary mapping from parameter names to values for
                         all parameters that take a fixed value.
        dumpSuffix: a string suffix to be added to the dump name.
        ---------
        Returns: a pair (dumpName, paramValuesDict), where
                 dumpName is a string name for dumps from the present parameter set
                 and paramValuesDict is a dictionary from parameter names to values
                 from the present parameter set.
        """
        
        #Initialize a dictionary mapping parameter names to values
        paramValuesDict = dict()
        
        #Update the dictionary with values of variable parameters
        for (varNames, paramValues) in variableParamNamesAndValuesList:
            paramValuesDict.update(dict(zip(varNames, paramValues)))
        
        #If there are no variable parameters, use a default dump name
        if len(paramValuesDict) == 0:
            dumpName = "dump"
        #Otherwise, create a dump name with the names and values of variable parameters            
        else:    
            dumpName = "_".join(["%s=%s" % pair for pair in sorted(iteritems(paramValuesDict))])
        #Add the suffix to the dump name
        dumpName += dumpSuffix
        
        #Add the fixed parameter values to the dictionary
        paramValuesDict.update(fixedParamsDict)
        
        return dumpName, paramValuesDict