import pandas as pd
import numpy as np
import os
import multiprocessing
from argparse import ArgumentParser
from exemplar_iotools import SettingsParser
from exemplar_analysistools import AnalysisTools
from exemplar_iotools import IOTools
from exemplar_representations import Agent
from builtins import range


class ExemplarModel:
    """A container in which agents are simulated.
    Specifies the constraints on the exemplar space and defines word frequency bins,
    but not the cognitive properties of exemplar systems.
    Contains high-level methods used for running simulations.
    """
    
    def __init__(self, spaceMin=-10, spaceMax=10, spacePoints=201, freqBins={"LF":[1,2,3],"MLF":[4,5,6],"MHF":[7,8,9],"HF":[10,11,12]}):
        """Initialize the parameters of the model. Do not load exemplars or run.
        =========
        Arguments
        ---------
        spaceMin: the minimum exemplar value permitted within a valid space.
        spaceMax: the maximum exemplar value permitted within a valid space.
        spacePoints: the number of points in the exemplar space (granularity).
        freqBins: a dictionary mapping from frequency classes ('LF', 'MLF', 'MHF', 'HF')
                  to the word frequencies they cover.
        """
        
        #Helper tool classes
        self.tools = AnalysisTools(self)
        self.io = IOTools(self)
        
        #Define the exemplar space
        self.spaceValues, self.spaceStep = np.linspace(start=spaceMin, stop=spaceMax, num=spacePoints, retstep=True)
        self.spaceValues = np.array([round(pos, 2) for pos in self.spaceValues]) 
        self.spaceIndices = {self.spaceValues[i] : i for i in range(spacePoints)}       
        self.spaceMin = spaceMin
        self.spaceMax = spaceMax
        self.spacePoints = spacePoints
        
        #Global properties
        self.freqBins = freqBins #maps from frequency class to frequency values
        self.agents = list() #will be converted to numpy array in postLoadSetup
        self.agentStore = dict() #maps from agent names to agents
        

    def addAgent(self, agent):
        """Add an agent to the store of agents.
        =========
        Arguments
        ---------
        agent: an Agent object.
        """

        self.agents.append(agent)
        self.agentStore[agent.name] = agent
        
        
    def setAgentProbabilities(self):
        """Set the agent selection probabilities."""
        
        #Create a uniform distribution over agents
        self.numAgents = len(self.agents)
        self.agents = np.array(self.agents)
        self.agentProbabilities = np.ones(self.numAgents) / self.numAgents
        
        
    def getAgent(self, agentName):
        """Get the agent with the given name.
        =========
        Arguments
        ---------
        agentName: a string specifying the name of an agent.
        =========
        Return: an Agent object.
        """
        
        agent = self.agentStore[agentName]
        
        return agent
        
   
    def chooseAgent(self):
        """Choose an agent at random and return it.
        =========
        Return: an Agent object.
        """
        
        #If there is only one agent, return that agent.
        if self.numAgents == 1:
            return self.agents[0]
            
        #Otherwise, choose at random following the (uniform) distribution in agentProbabilities
        agent = np.random.choice(self.agents, p=self.agentProbabilities)
        
        return agent 
                
               
    def postLoadSetup(self):
        """Perform setup after loading data by creating probability distributions for:
        - word type selection;
        - category selection;
        - listener selection;
        - agent (speaker) selection.
        """
        
        for agent in self.agents:
        
            #Within each category, set word type selection probabilities (based on frequency)
            for category in agent.categories:                
                category.setWordProbabilities()
                        
            #Set category selection probabilities
            agent.setCategoryProbabilities()
            
            #Set listener selection probabilities
            agent.setListenerProbabilities()
            
            #Set recognition threshold and storage threshold by frequency
            for freq in range(1, agent.maxFreq+1):
                #Recognition threshold linearly decreasing with frequency, within [0,1]
                agent.recogThreshold[freq] = min(1, max(0, agent.recogThresholdRatio - agent.recogThresholdFreqDiff * (2 * (freq - 1) / float(agent.maxFreq - 1) - 1)))                        
                #Storage threshold linearly increasing with frequency, within [0,1]
                agent.storeThreshold[freq] = min(1, max(0, agent.storeThresholdProportion + agent.storeThresholdFreqDiff * (2 * (freq - 1) / float(agent.maxFreq - 1) - 1)))
                
        #Set agent selection probabilities
        self.setAgentProbabilities()
    
    
    def runModel(self, iterations, dumpEvery=None, dumpStates=False, dumpsDir="Dumps", dumpSummary=True, dumpName="dump", logsDir="Logs"):
        """Run the model for a specified number of iterations.
        =========
        Arguments
        ---------
        iterations: the number of iterations to run the model for.
        dumpEvery: the number of iterations between logs of the system (summaries 
                   or states). A value of None indicates no logging.
        dumpStates: a Boolean flag indicating whether to dump states of the system,
                    at the level of individual exemplars.
        dumpsDir: a string indicating the directory to store system state dumps.
        dumpSummary: a Boolean flag indicating whether to dump summaries of the
                     system, at the level of categories and frequency classes.
        dumpName: a string indicating the file prefix for dumps.
        logsDir: a string indicating the directory to store summary logs.
        """
        
        #If the model is creating logs, create an initial log
        if dumpEvery is not None:
            #Create a log of summary statistics
            summaryDict = self.tools.makeSummaryDict()
            summaryDict = self.tools.updateSummaryDict(summaryDict, 0)
            #If the model is logging states as well as summaries, dump the first state 
            if dumpStates:
                for agent in self.agents:
                    self.io.dumpData(agent=agent, filename="%s/%s_%s_iter0.json" % (dumpsDir, dumpName, agent.name))
        
        #Commence iterations. Each iteration corresponds to a single utterance.
        for i in range(1,iterations+1):
            
            #Choose the speaker and listener
            speaker = self.chooseAgent()
            listeners = speaker.chooseListeners()
            
            #The speaker produces a token          
            token = speaker.produce()
            
            #Each listener perceives the token
            for listener in listeners:
                listener.perceive(token)
            
            #Log summary information and dump states at regular intervals, if required
            if dumpEvery is not None and i % dumpEvery == 0:
                summaryDict = self.tools.updateSummaryDict(summaryDict, i)
                if dumpStates:
                    for agent in self.agents:
                        self.io.dumpData(agent=agent, filename="%s/%s_%s_iter%s.json" % (dumpsDir, dumpName, agent.name, i))
                      
        #If the model is creating logs, make the final summary log and then dump the summary file
        if dumpEvery is not None:
            self.summaryDataFrame = self.tools.makeSummaryDataFrame(summaryDict)
            if dumpSummary:
                self.io.dumpDataFrame(summaryDict, path=logsDir, fileName="%s_summary.csv" % dumpName)
                
    
    def runMultiple(self, totalRuns, iterations, dumpEvery=None, dumpStates=False, dumpsDir="Dumps", dumpName="dump", shuffleWords=True, shuffleFrequencies=False, logsDir="Logs"):
        """Run the model from starting point multiple times.
        =========
        Arguments
        ---------
        totalRuns: the number of times to repeat the model run from the initial
                   conditions, through all iterations.
        iterations: the number of iterations of the model.
        dumpEvery: the number of iterations between logs of the system (summaries 
                   or states). A value of None indicates no logging.
        dumpStates: a Boolean flag indicating whether to dump states of the system,
                    at the level of individual exemplars.
        dumpsDir: a string indicating the directory to store system state dumps.
        dumpSummary: a Boolean flag indicating whether to dump summaries of the
                     system, at the level of categories and frequency classes.
        dumpName: a string indicating the file prefix for dumps.
        shuffleWords: a Boolean flag indicating whether to shuffle exemplars across
                      word types within a frequency class on different runs.
        shuffleFrequencies: a Boolean flag indicating whether to shuffle the 
                            frequencies of word types within the same category
                            on different runs.
        logsDir: a string indicating the directory to store summary logs.
        """
        
        #Storage of summary data frames
        summaryDataFramesList = list()
        
        for runNumber in range(totalRuns):
            
            #Reload data
            for agent in self.agents:
                self.io.loadData(agent, self.io.initDataFilename, shuffleWords=shuffleWords, clearExemplars=True, shuffleFrequencies=shuffleFrequencies)
            
            #Run model
            self.runModel(iterations, dumpSummary=False, dumpEvery=dumpEvery, dumpStates=dumpStates, dumpsDir=dumpsDir, dumpName="%s_run%s" % (dumpName, runNumber))
            
            #Add run number to summary data frame and append to list.
            self.summaryDataFrame['run'] = runNumber
            summaryDataFramesList.append(self.summaryDataFrame)
        
        #Concatenate different summary data frames and log
        self.summaryDataFrame = pd.concat(summaryDataFramesList, ignore_index=True)
        self.io.dumpDataFrame(self.summaryDataFrame, path=logsDir, filename="%s_summary.csv" % dumpName)             
        
    
def main(settingsFilename="settings.json", totalPools=1, poolNum=0, processesPerPool=1, logsDir="Logs", dumpStates=False, dumpsDir="Dumps"):
    """Initiate and run a model using the settings provided in a file.
    =========
    Arguments
    ---------
    settingsFilename: a path to the settings json file.
    totalPools: the number of pools to split model repeats across.
    poolNum: the current pool number.
    processesPerPool: the number of processes to split model runs across within
                      a pool.
    logsDir: a string indicating the directory to store summary logs.                      
    dumpStates: a Boolean flag indicating whether to dump states of the system,
                at the level of individual exemplars.
    dumpsDir: a string indicating the directory to store system state dumps.
    """
        
    #Create the directories for logs and dumps, if required
    if not os.path.exists(logsDir):
        os.makedirs(logsDir)
    if dumpStates and not os.path.exists(dumpsDir):
        os.makedirs(dumpsDir)
        
    #Parse the settings file
    settings = SettingsParser(settingsFilename)
    
    #Extract the meta settings (properties of the model)
    #and the parameter combinations (properties of the agents)
    metaSettings = (settings["freqBins"], settings["dataPath"], settings["runsPerSetting"], settings["iterations"], settings["dumpEvery"], settings["spaceMin"], settings["spaceMax"], settings["spacePoints"], settings["shuffleWords"], settings["shuffleFrequencies"], logsDir, dumpStates, dumpsDir)
    allParamCombos = settings.paramGenerator()
    
    #If no parallel processing is required, launch a series of models in serial, one per parameter combination
    if totalPools * processesPerPool == 1:
        for dumpName, initSettings in allParamCombos:
            launchModel(((dumpName, initSettings), metaSettings))
            
    else:      
        #If parallel processing is required, identify the parameter combinations for the current pool
        paramCombosThisPool = int(settings.totalParamCombos / totalPools)
        if poolNum < settings.totalParamCombos - (paramCombosThisPool * totalPools):
            paramCombosThisPool += 1
            
        #Skip the parameter combinations which are handled by other pools
        for i in range(int(settings.totalParamCombos / totalPools) * poolNum + min(poolNum, settings.totalParamCombos % totalPools)):
            skippedParamCombo = next(allParamCombos)
            
        #Extract the parameter combinations for the current pool
        paramCombos = [next(allParamCombos) for i in range(paramCombosThisPool)]    
        
        #Launch processes for this pool
        pool = multiprocessing.Pool(processes=processesPerPool)
        pool.map(launchModel, zip(paramCombos, [metaSettings]*paramCombosThisPool))
        pool.close()
        pool.join()   
    

def launchModel(paramCombos_metaSettings):
    """Create a model with the given settings and run it.
    =========
    Arguments
    ---------
    paramCombos_metaSettings: a pair (paramCombos, metaSettings) [see below]
    paramCombos: a pair (dumpName, initSettings) [see below]
    dumpName: a string identifier for the current run, used to label dumped logs.
    initSettings: a dictionary specifying the parameter settings for the current runs.
    metaSettings: a tuple (freqBins, dataPath, runsPerSetting, iterations, dumpEvery, 
                           spaceMin, spaceMax, spacePoints, shuffleWords, 
                           shuffleFrequencies, logsDir, dumpStates, dumpsDir) [see below]
    freqBins: a dictionary mapping from frequency classes ('LF', 'MLF', 'MHF', 'HF')
                to the word frequencies they cover.
    dataPath: the path to the initial data json.
    runsPerSetting: the number of runs to perform with the present parameter settings.
    iterations: the number of iterations to run the model for.
    dumpEvery: the number of iterations between logs of the system (summaries 
                or states). A value of None indicates no logging.
    spaceMin: the minimum exemplar value permitted within a valid space.
    spaceMax: the maximum exemplar value permitted within a valid space.
    spacePoints: the number of points in the exemplar space (granularity).                
    shuffleWords: a Boolean flag indicating whether to shuffle exemplars across
                    word types within a frequency class on different runs.
    shuffleFrequencies: a Boolean flag indicating whether to shuffle the 
                        frequencies of word types within the same category
                        on different runs.    
    logsDir: a string indicating the directory to store summary logs.
    dumpStates: a Boolean flag indicating whether to dump states of the system,
                at the level of individual exemplars.
    dumpsDir: a string indicating the directory to store system state dumps.
    """
    
    #Unpack arguments
    (paramCombos, metaSettings) = paramCombos_metaSettings
    (dumpName, initSettings) = paramCombos
    (freqBins, dataPath, runsPerSetting, iterations, dumpEvery, spaceMin, spaceMax, spacePoints, shuffleWords, shuffleFrequencies, logsDir, dumpStates, dumpsDir) = metaSettings
    
    #Create model
    model = ExemplarModel(spaceMin=spaceMin, spaceMax=spaceMax, spacePoints=spacePoints, freqBins=freqBins)
    
    #Create agent
    agent = Agent(model, "agent0", **initSettings)
    model.addAgent(agent)
    
    #Get category and word properties for the agent (note that shuffling is irrelevant at this point, so is disabled)
    model.io.loadData(agent, dataFilename=dataPath, shuffleWords=False, shuffleFrequencies=False, makeAdjustments=True)
    model.postLoadSetup()
    
    #Run the model (shuffling before running as required)
    model.runMultiple(runsPerSetting, iterations, dumpEvery=dumpEvery, dumpStates=dumpStates, dumpsDir=dumpsDir, dumpName=dumpName, shuffleWords=shuffleWords, shuffleFrequencies=shuffleFrequencies, logsDir=logsDir)  
    
    
if __name__ == '__main__':
    
    #Arguments to read from command line
    parser=ArgumentParser()
    parser.add_argument('--pools', help="Total pools of processes to launch",
        type=int, default=1)
    parser.add_argument('--pool-number', help="Number of current pool",
        type=int, default=0)
    parser.add_argument('--processes-per-pool', help="Number of processes per pool", 
        type=int, default=1)
    parser.add_argument('--settings-path', help="Path to settings json file", 
        type=str, default="settings.json")   
    parser.add_argument('--logs-path', help="Path to directory for writing summary logs", 
        type=str, default="Logs")   
    parser.add_argument('--dump-exemplars', help="Provide this flag to dump the exemplar clouds (in addition to category-level summaries)",
        action='store_true')   
    parser.add_argument('--dumps-path', help="Path to directory for saving exemplar dump states",
        type=str, default="Dumps")              
    args= parser.parse_args()
    
    #Initiate a model when called from the command line
    main(settingsFilename=args.settings_path, totalPools=args.pools, poolNum=args.pool_number, processesPerPool=args.processes_per_pool, logsDir=args.logs_path, dumpStates=args.dump_exemplars, dumpsDir=args.dumps_path)