import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import os
from argparse import ArgumentParser
from exemplar_iotools import SettingsParser
from exemplar_model import ExemplarModel
from exemplar_representations import Agent
from builtins import range


class PlotTools:
    """A container for methods involving the plotting of data from dumps of
    the exemplar system state.
    """
        
    def __init__(self, model):
        """Creates a class containing tools for plotting model outputs.
        =========
        Arguments
        ---------
        model: an ExemplarModel object with the same properties as that in which 
               the output was obtained.
        """
          
        self.model = model
        
        
    def plotDistributions(self, agent, categoriesToPlot, categoryColors, divider=1.0, ymax=None, plotOnlyFreqs=False, **otherArgs):
        """Plot the distributions of exemplars in each category, unsmoothed.
        Can instead plot high- and low-frequency sub-distributions, via plotOnlyFreqs.
        =========
        Arguments
        ---------
        agent: the Agent object containing the exemplars to plot.
        categoriesToPlot: a list of names of categories within the agent to plot. 
        categoryColors: a dictionary from category names to matplotlib color letters.     
        divider: a normalizing float to divide the counts by before plotting.
        ymax: the maximum value for the y-axis, or None. If None, a value will be 
              automatically computed so as to contain all of the data in the plot.
        plotOnlyFreqs: a Boolean flag indicating whether to plot only the frequency-based
                       sub-distributions (True) or the combined distributions over 
                       exemplars of all frequencies (False).
        otherArgs: a dictionary of leftover keyword arguments from the feeder method,
                   not required for this method.
        """
        
        #Initialize the list of Line2D objects to plot in the category legend
        categoryLines = list()

        #If plotting frequency-based sub-distributions, get those sub-distributions
        if plotOnlyFreqs:
            hfDists = self.model.tools.getFilteredExemplarHistograms(agent, self.model.tools.hfFilter)
            lfDists = self.model.tools.getFilteredExemplarHistograms(agent, self.model.tools.lfFilter) 
        #Otherwise, get all exemplars from each category    
        else:
            allDists = self.model.tools.getFilteredExemplarHistograms(agent, self.model.tools.identFilter)
        
        #Iterate through the categories
        for categoryName in categoriesToPlot:
            
            #Plot the distribution (or subdistributions)
            if plotOnlyFreqs:
                plt.plot(self.model.spaceValues, hfDists[categoryName] / divider, '%s--' % categoryColors[categoryName])
                plt.plot(self.model.spaceValues, lfDists[categoryName] / divider, '%s-' % categoryColors[categoryName])
            else:
                plt.plot(self.model.spaceValues, allDists[categoryName] / divider, '%s-' % categoryColors[categoryName])
            
            #Add the category to the legend
            categoryLines.append(mlines.Line2D([], [], color=categoryColors[categoryName], label=categoryName))
        
        #If no ymax is specified, choose the smallest multiple of 20 that will include all y values
        if ymax is None:
            ymax = 0
            for category in categoriesToPlot:
                if plotOnlyFreqs:
                    ymax = max(ymax, hfDists[category].max(), lfDists[category].max())
                else:
                    ymax = max(ymax, allDists[category].max())
            ymax /= divider
            ymax = (int(ymax / 20) + 1) * 20
        
        #Set the limits of the y axis                
        plt.axis([self.model.spaceMin, self.model.spaceMax, 0, ymax])
        
        #Label the axes
        plt.xlabel("Exemplar value")
        plt.ylabel("Number of exemplars")           
        
        #If plotting frequency sub-distributions, add a legend for frequency class
        if plotOnlyFreqs:
            dashed_line = mlines.Line2D([], [], linestyle='dashed', color="black", label="HF only")
            solid_line = mlines.Line2D([], [], linestyle='solid', color="black", label="LF only")
            freq_legend = plt.legend([dashed_line, solid_line], ["HF only", "LF only"], loc=2)
            plt.gca().add_artist(freq_legend)
            
        #Add a legend for category
        plt.legend(categoryLines, categoriesToPlot, loc=1)             
        
        
    def plotActivations(self, agent, categoriesToPlot, categoryColors, divider=1.0, ymax=None, plotOnlyFreqs=False, addExemplars=True, **otherArgs):
        """Plots the activation fields of each category.
        Can instead plot high- and low-frequency activation fields, via plotOnlyFreqs.
        Can also plot the exemplar locations in memory, via addExemplars.
        =========
        Arguments
        ---------
        agent: the Agent object containing the exemplars to plot.
        categoriesToPlot: a list of names of categories within the agent to plot. 
        categoryColors: a dictionary from category names to matplotlib color letters.     
        divider: a normalizing float to divide the counts by before plotting.
        ymax: the maximum value for the y-axis, or None. If None, a value will be 
              automatically computed so as to contain all of the data in the plot.
        plotOnlyFreqs: a Boolean flag indicating whether to plot only the activation
                       fields of the frequency-based sub-distributions (True) or the
                       combined field over exemplars of all frequencies (False).
        addExemplars: a Boolean flag indicating whether to add a rug at the bottom
                      of the plot representing actual exemplar locations.
        otherArgs: a dictionary of leftover keyword arguments from the feeder method,
                   not required for this method.
        """
        
        #Initialize the list of Line2D objects to plot in the category legend        
        categoryLines = list()
        
        #If plotting frequency-based sub-distributions, get the activation fields of those sub-distributions        
        if plotOnlyFreqs:
            hfActs = self.model.tools.getFilteredActivationFields(agent, self.model.tools.hfFilter)
            lfActs = self.model.tools.getFilteredActivationFields(agent, self.model.tools.lfFilter)
        #Otherwise, get the activation field across all exemplars from each category            
        else:
            allActs = {category.name: category.actSpace for category in agent.categories}
        
        #Iterate through the categories        
        for categoryName in categoriesToPlot:
            
            #Plot the activation field(s)
            if plotOnlyFreqs:
                plt.plot(self.model.spaceValues, hfActs[categoryName] / divider, '%s--' % categoryColors[categoryName])
                plt.plot(self.model.spaceValues, lfActs[categoryName] / divider, '%s-' % categoryColors[categoryName])
            else:
                plt.plot(self.model.spaceValues, allActs[categoryName] / divider, '%s-' % categoryColors[categoryName])
            
            #Get all exemplars corresponding to the field(s) and add them to the bottom of the plot
            if addExemplars:
                categoryExemplars = self.model.tools.getExemplarValues(self.model.tools.getFilteredWordCloud(agent.getCategory(categoryName), lambda word: True))
                if plotOnlyFreqs:
                    plt.scatter(categoryExemplars, 0.4 * np.ones(categoryExemplars.shape), alpha=0.05 / divider, color=categoryColors[categoryName], label="_nolegend_")
                else:
                    plt.scatter(categoryExemplars, 2 * np.ones(categoryExemplars.shape), alpha=0.02 / divider, color=categoryColors[categoryName], label="_nolegend_")
            
            #Add the category to the legend
            categoryLines.append(mlines.Line2D([], [], color=categoryColors[categoryName], label=categoryName))
        
        #If no ymax is specified, choose the smallest multiple of 100 that will include all y values        
        if ymax is None:
            ymax = 0
            for category in categoriesToPlot:
                if plotOnlyFreqs:
                    ymax = max(ymax, hfActs[category].max(), lfActs[category].max())
                else:
                    ymax = max(ymax, allActs[category].max())
            ymax /= divider
            ymax = (int(ymax / 100) + 1) * 100
            
        #Set the limits of the y-axis    
        plt.axis([self.model.spaceMin, self.model.spaceMax, 0, ymax])   
        
        #Set the axis labels
        plt.xlabel("Exemplar value")
        plt.ylabel("Activation")          
        
        #If plotting frequency sub-distributions, add a legend for frequency class
        if plotOnlyFreqs:
            dashed_line = mlines.Line2D([], [], linestyle='dashed', color="black", label="HF only")
            solid_line = mlines.Line2D([], [], linestyle='solid', color="black", label="LF only")
            freq_legend = plt.legend([dashed_line, solid_line], ["HF only", "LF only"], loc=2)
            plt.gca().add_artist(freq_legend)
            
        #Add a legend for category
        plt.legend(categoryLines, categoriesToPlot, loc=1)                
                
        
    def makePlots(self, plotWhat, agent, categoriesToPlot=None, categoryColors=None, dirName="Plots", filename="plot", figSizePixels=(800,600), figDPI=96, **kwargs):
        """Make a plot of a given type for given categories, and save it as a PNG.
        =========
        Arguments
        ---------
        plotWhat: the string name of the plot to make. Either "Activations" or "Distributions".
        agent: the Agent object containing the exemplars to plot.
        categoriesToPlot: a list of the string names of categories to plot, or None.
                          If None, all categories in the agent will be plotted.
        categoryColors: a dictionary mapping from category names to matplotlib
                        color letters, or None. If None, colors will be assigned
                        to up to 7 categories.
        dirName: the name of the directory in which to save the plot.
        filename: the name of the plot, excluding extension.
        figSizePixels: a pair specifying the width and height of the plot, in pixels.
        figDPI: the dots-per-inch value for the figure.
        kwargs: a dictionary of keyword arguments to be fed to the plotting method.
        """
        
        #If no categories are provided, plot all categories in the agent
        if categoriesToPlot is None:
            categoriesToPlot = agent.categoryStore.keys()
        
        #If no category colors are provided, assign up to 7.
        #Raise an exception if more than 7 categories are to be plotted.
        if categoryColors is None:
            if len(categoriesToPlot) > 7:
                raise Exception("ERROR: Only 7 categories can be plotted at once.")
            colors = ['b', 'r', 'g', 'c', 'm', 'y', 'k']
            categoryColors = dict(zip(categoriesToPlot, colors))
        
        #Change figure size
        fig = plt.gcf()
        fig.set_size_inches(float(figSizePixels[0])/figDPI, float(figSizePixels[1])/figDPI)
        
        #Create the plot.
        #Raise an exception if an unknown plot type is asked for.
        if plotWhat == "Activations":
            self.plotActivations(agent, categoriesToPlot, categoryColors, **kwargs)  
        elif plotWhat == "Distributions":
            self.plotDistributions(agent, categoriesToPlot, categoryColors, **kwargs)
        else:
            raise Exception("ERROR: No plotting function of type %s." % plotWhat)

        #Save the plot and close the figure.
        plt.savefig("%s/%s_%s.png" % (dirName, plotWhat, filename), dpi=figDPI)
        plt.close()   
            
            
            
def main(settingsFilename="settings.json", dumpsDir="Dumps", plotsDir="Plots", plotWhat="Activations", plotRunsSeparately=False, **plotArgs):
    """Create plots from simulations whose states have been dumped.
    =========
    Arguments
    ---------
    settingsFilename: the name of the settings JSON file used for the simulations.
    dumpsDir: the name of the directory in which the states have been dumped.
    plotsDir: the name of the directory in which plots should be saved.
    plotWhat: a string indicating the kind of plot(s) that should be generated;
              "Activations", "Distributions", or "Both".
    plotRunsSeparately: a Boolean flag indicating whether to plot different runs
                        with the same settings separately, or whether to combine
                        them into a plot of average (expected) state.
    plotArgs: a dictionary of keyword arguments to be fed to the plotting method.
    """
    
    #Create the plots directory if it doesn't already exist
    if not os.path.exists(plotsDir):
        os.makedirs(plotsDir)
        
    #Convert plotWhat to a list which can be iterated over
    if plotWhat == "Both":
        plotWhat = ["Activations", "Distributions"]
    else:
        plotWhat = [plotWhat]
        
    #Read settings from the settings file
    settings = SettingsParser(settingsFilename)
    freqBins = settings["freqBins"]
    repeats = settings["repeatsPerRun"]
    runs = settings["runsPerSetting"]
    spaceMin = settings["spaceMin"]
    spaceMax = settings["spaceMax"]
    spacePoints = settings["spacePoints"]
    dumpIters = range(0, settings["iterations"]+1, settings["dumpEvery"])
    paramCombos = settings.paramGenerator()    
    
    #Compute a normalizing divider to use when plotting separate runs together, to get averages
    divider = float(runs * repeats)
    
    #If plotting runs separately, load and plot each file generated by the settings
    if plotRunsSeparately:
        for dumpName, initSettings in paramCombos:
            for run in range(runs):
                for iteration in dumpIters:
                    
                    #Reconstruct the filenames based on parameter combinations
                    filename = "%s_run%s_agent0_iter%s" % (dumpName, run, iteration)
                    
                    #Create a container ExemplarModel for the data
                    model = ExemplarModel(spaceMin=spaceMin, spaceMax=spaceMax, spacePoints=spacePoints, freqBins=freqBins)
                    
                    #Create a container Agent for the model, and initialize it with the dump data
                    agent = Agent(model, "agent0", **initSettings)
                    model.addAgent(agent)
                    model.io.loadData(agent, dataFilename="%s/%s.json" % (dumpsDir, filename), shuffleWords=False, makeAdjustments=False)
                    
                    #Make the required plots for the given run
                    plotter = PlotTools(model)
                    for plotThis in plotWhat:
                        plotter.makePlots(plotThis, agent, dirName=plotsDir, filename=filename.replace("_agent0",""), **plotArgs)
                 
    #If plotting runs together, load all files generated by the same parameter settings
    #together, and plot their average
    else:
        for dumpName, initSettings in paramCombos:
            
            #Restrict to the first repeat, since the mass loading method will
            #generalize to include all repeats
            if repeats == 1 or dumpName.endswith("_repeatNum=1"):
                dumpName = dumpName.replace("_repeatNum=1", "")
                for iteration in dumpIters:
                    
                    #Create a filename regular expression for all runs with the same parameter settings
                    repeatPart = ""
                    runPart = ""
                    if repeats > 1:
                        repeatPart = "_repeatNum=\d+"
                    if runs > 1:
                        runPart = "_run\d+"
                    fileNamePattern = "%s%s%s_agent0_iter%s" % (dumpName, repeatPart, runPart, iteration)
                    
                    #Create a container ExemplarModel for the data
                    model = ExemplarModel(spaceMin=spaceMin, spaceMax=spaceMax, spacePoints=spacePoints, freqBins=freqBins)
                    
                    #Create a container Agent for the model, and initialize it with the dump data
                    #for all filenames matching the regular expression, i.e. all runs with the given parameters
                    agent = Agent(model, "agent0", **initSettings)
                    model.addAgent(agent)
                    model.io.loadMultiple(agent, dataPath=dumpsDir, fileNameSuffix=fileNamePattern)
                    
                    #Make the required plots for the given run
                    plotter = PlotTools(model)
                    for plotThis in plotWhat:
                        plotter.makePlots(plotThis, agent, dirName=plotsDir, filename="%s_iter%s" % (dumpName, iteration), divider=divider, **plotArgs)
        
        
if __name__ == '__main__':
    
    #Arguments to read from command line
    parser=ArgumentParser()
    parser.add_argument('--settings-path', help="Path to settings json file", 
        type=str, default="settings.json")   
    parser.add_argument('--dumps-path', help="Path to directory where exemplar dump states are saved", 
        type=str, default="Dumps")   
    parser.add_argument('--plots-path', help="Path to directory for saving plots", 
        type=str, default="Plots")
    parser.add_argument('--plot', help="Plot Activations (default), Distributions, or Both", 
        type=str, default="Activations")            
    parser.add_argument('--separate-runs', help="Provide this flag to plot runs separately",
        action='store_true')                 
    parser.add_argument('--plot-freqs', help="Provide this flag to plot HF and LF sub-distributions",
        action='store_true')         
    parser.add_argument('--no-exemplars', help="Provide this flag to remove exemplars from an Activation plot",
        action='store_false')    
    parser.add_argument('--ymax', help="The maximum value to be plotted on the y-axis",
        type=int, default=None)                  
    args= parser.parse_args()
    
    #Initiate plotting when called from the command line
    main(settingsFilename=args.settings_path, dumpsDir=args.dumps_path, plotsDir=args.plots_path, plotWhat=args.plot, plotRunsSeparately=args.separate_runs, plotOnlyFreqs=args.plot_freqs, addExemplars=args.no_exemplars, ymax=args.ymax)