import numpy as np
import random
from builtins import range


class Agent:
    """A representation of an agent, who has their own set of categories, words, 
    and exemplar system, defined by their own parameters.
    Contains methods for production and perception.
    """
    
    def __init__(self, model, name, sigma=0.6, mu=1.9, beta=0.10, iota=0.5, alpha=0.3, delta=0.25, deltaDiff=0, tau=0.10, tauDiff=0, epsilon=0, chi=1):
        """Initialize the key attributes of the agent.
        =========
        Arguments
        ---------
        model: the ExemplarModel object in which the agent exists.
        name: the unique string name used to identify the agent.
        sigma: the initial category width.
        mu: the initial category separation.
        beta: the size of the bias.
        iota: the degree of imprecision.
        alpha: the size of the activation window.
        delta: the discriminability threshold for median-frequency words.
        deltaDiff: the difference between the discriminability threshold for median-
                   frequency words and for high/low-frequency words.
        tau: the typicality threshold for median-frequency words.
        tauDiff: the difference between the typicality threshold for median-
                 frequency words and for high/low-frequency words.
        epsilon: the size of the entrenchment window.
        chi: the degree of nonword competition in recognition.
        """
        
        #Backpointer
        self.model = model
        
        #Identification
        self.name = name

        #Parameters
        self.perWindowSD = alpha
        self.proWindowSD = epsilon
        self.noiseSD = iota 
        self.biasSF = beta
        self.initCategoryWidthSD = sigma
        self.initCategoryAdjustSize = mu 
        self.nonwordCompetition = chi  
        
        self.recogThresholdRatio = delta
        self.recogThresholdFreqDiff = deltaDiff
        self.recogThreshold = dict() #will be given a value in model.postLoadSetup
        
        self.storeThresholdProportion = tau
        self.storeThresholdFreqDiff = tauDiff
        self.storeThreshold = dict() #will be given a value in model.postLoadSetup             
        
        #Create a lookup table for pre-defined activation windows in perception, for efficiency       
        if self.perWindowSD > 0:
            perWindowExp = 1.0 / (2 * self.perWindowSD ** 2)
            self.perWindows = {pos : np.exp(-perWindowExp * (self.model.spaceValues-pos)**2) for pos in self.model.spaceValues}
        else:
            self.perWindows = {pos : np.array([0]*self.model.spaceIndices[pos] + [1] + [0]*(self.model.spacePoints-self.model.spaceIndices[pos]-1)) for pos in self.model.spaceValues}
        
        #Create a lookup table for pre-defined entrenchment windows in perception, for efficiency
        if self.proWindowSD > 0:
            proWindowExp = 1.0 / (2 * self.proWindowSD ** 2)
            self.proWindows = {pos : np.exp(-proWindowExp * (self.model.spaceValues-pos)**2) for pos in self.model.spaceValues}
            self.convProWindows = {pos : pos * np.exp(-proWindowExp * (self.model.spaceValues-pos)**2) for pos in self.model.spaceValues}
        else:
            self.proWindows = {pos : np.array([0]*self.model.spaceIndices[pos] + [1] + [0]*(self.model.spacePoints-self.model.spaceIndices[pos]-1)) for pos in self.model.spaceValues}
            self.convProWindows = {pos : np.array([0]*self.model.spaceIndices[pos] + [pos] + [0]*(self.model.spacePoints-self.model.spaceIndices[pos]-1)) for pos in self.model.spaceValues}
        
        #Initialize global properties                 
        self.categories = list()
        self.categoryStore = dict()
        self.numCategories = 0
        self.maxFreq = 0 
        self.listenerProbabilities = dict() # will be filled in via agent.setListenerProbabilities during model.postLoadSetup
        
    
    def addCategory(self, categoryCloud):
        """Add a category cloud to the store of categories.
        =========
        Arguments
        ---------
        categoryCloud: a CategoryCloud object.
        """
        
        self.categories.append(categoryCloud)
        self.categoryStore[categoryCloud.name] = categoryCloud
    
            
    def setCategoryProbabilities(self):
        """Set the category selection probabilities, based on the sum of word
        frequencies within each category.
        """
        
        #Category selection probabilities are determined by the sum of word frequencies within each category
        self.numCategories = len(self.categories)
        self.categories = np.array(self.categories)
        self.categoryProbabilities = np.array([category.totalWordFrequency for category in self.categories])
        self.categoryProbabilities /= np.sum(self.categoryProbabilities)
        
        
    def getCategory(self, categoryName):
        """Get the category with the given name.
        =========
        Arguments
        ---------
        categoryName: a string representation of the category name.
        =========
        Return: the CategoryCloud object which has the given name.
        """
        
        category = self.categoryStore[categoryName]
        
        return category
        
        
    def getLandscapeValue(self, landscape, point):
        """Get the value in a landscape (e.g. activation field) at a particular 
        point.
        =========
        Arguments
        ---------
        landscape: an array of values over the exemplar space, e.g. representing
                   an activation field.
        point: the acoustic value at which to return the landscape value.
        =========
        Return: a float representing the value of the landscape at the point.
        """
        
        #Get the array index to the point
        index = self.model.spaceIndices[point]
        
        return landscape[index]  
        
               
    def chooseCategory(self):
        """Choose a category to produce and return it.
        =========
        Return: a CategoryCloud object.
        """
        
        #Categories are chosen following the distribution from categoryProbabilities
        category = np.random.choice(self.categories, p=self.categoryProbabilities)
        
        return category
        
    
    def getWordsFromFrame(self, frame):
        """Get a list of the words that are consistent with a given frame,
        including possible nonwords.
        =========
        Arguments
        ---------
        frame: a string representation of the phonological frame.
        =========
        Return: a list of WordType objects that are consistent with the frame.
        """
        
        words = [category.getWordType(frame) for category in self.categories]
        
        return words
        
        
    def setListenerProbabilities(self):
        """Create the dictionary of probabilities with which the agent will speak
        to other agents."""
        
        #By default, the listener probabilities are uniform
        for otherAgent in self.model.agents:
            if otherAgent != self:
                self.listenerProbabilities[otherAgent.name] = 1.0 / (self.model.numAgents-1)
        
        
    def chooseListeners(self):
        """Chooses the listeners for a given utterance. An agent is always their
        own listener. If there are multiple agents, a single extra listener is 
        chosen from the other agents, according to agent.listenerProbabilities.
        =========
        Return: a list of Agent objects to function as listeners.
        """
        
        #Single agent: speaker is listener
        if self.model.numAgents == 1:
            listeners = [self]
        
        #Two agents: both listening
        elif self.model.numAgents == 2:
            listeners = self.model.agents
        
        #More than two agents: speaker and one other listening, chosen following listenerProbabilities
        else:
            otherListener = self.model.getAgent(np.random.choice(self.listenerProbabilities.keys(), p=self.listenerProbabilities.values()))
            listeners = [self, otherListener]
        
        return listeners        

        
    def produce(self):
        """Choose a category and word to produce, pick an exemplar, make a copy,
        entrench it, add noise and possibly bias, discretize the value, and return
        the new exemplar.
        Note: productions are not allowed to leave the designated space. A production
        with a target outside the space will be moved back to the edge.
        =========
        Return: an Exemplar object, representing a produced token.
        """
        
        #Choose the production target
        category = self.chooseCategory()
        word = category.chooseWord()
        targetExemplar = word.chooseExemplar()
        
        #Create an Exemplar object
        exemplar = Exemplar(targetExemplar.value, word.frame, self)
        
        #Perform production processes for the exemplar
        self.entrenchExemplar(exemplar, word)
        self.addExemplarBias(exemplar, word)            
        self.addExemplarNoise(exemplar, word)
            
        #Discretize and bring exemplar back into space if it has left
        exemplar.discretize()
        exemplar.bringIntoSpace()
        
        return exemplar    
        
        
    def perceive(self, exemplar):
        """Categorize an incoming exemplar and attempt to store it.
        Nonword categorizations are not stored, following a condition check
        within the typicality evaluation.
        =========
        Arguments
        ---------
        exemplar: an Exemplar object, representing the token heard.
        """
        
        #Retrieve the perceived word, based on the acoustic value and frame
        perceivedWord = self.recognizeExemplar(exemplar)

        #Typicality evaluation. Nonwords automatically fail, so they can't be stored.
        isTypical = self.typicalityEvaluation(exemplar, perceivedWord)

        #Store the exemplar if it passes the typicality evaluation
        if isTypical:
            
            #Storage entails overwriting a random exemplar
            self.storeExemplar(perceivedWord, exemplar)
            
    
    def entrenchExemplar(self, exemplar, word):
        """Entrench a given exemplar. Assumed: value is discretized and in the exemplar store already.
        Note that the returned value is not discretized (in the chain, entrenchment happens first,
        then bias, then discretization upon perception).
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be entrenched.
        word: the WordType object for the exemplar.
        """
        
        #Only make a change if there is entrenchment
        if self.proWindowSD != 0:
            
            #Entrenchment is a center of mass calculation: 
            #the area under a convolution of the entrenchment window,
            #divided by the area under the entrenchment window
            C = self.getLandscapeValue(word.category.convProSpace, exemplar.value)
            P = self.agent.getLandscapeValue(self.category.proSpace, self.value)
            
            #Update exemplar value
            exemplar.value = C/P        
        
        
    def addExemplarBias(self, exemplar, word):
        """Add production bias to a given exemplar.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be biased.
        word: the WordType object for the exemplar.
        """
        
        #Only make a change if there is bias for the category
        if word.category.scaledBias is not None:
            
            #Update exemplar value
            exemplar.value += word.category.scaledBias        
    
    
    def addExemplarNoise(self, exemplar, word):
        """Add production noise to a given exemplar.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be biased.
        word: the WordType object for the exemplar.
        """
        
        #Only make a change if there is noise
        if self.noiseSD != 0:
            
            #Update exemplar value
            exemplar.value += np.random.normal(scale = self.noiseSD)
        
    
    def recognizeExemplar(self, exemplar):
        """Recognize an exemplar as an instance of a particular word, rather 
        than competing words. Combines identification with the discriminability 
        evaluation.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be recognized.
        =========
        Return: the WordType object that the exemplar is recognized to correspond to.
        """
        
        #Get the candidate words
        candidates = self.getWordsFromFrame(exemplar.wordFrame)
        
        #If any of the words have a recognition threshold of zero, choose randomly from among those words.
        thresholdZeroWords = [candidate for candidate in candidates if candidate.frequency != 0 and self.recogThreshold[candidate.frequency] == 0]
        if len(thresholdZeroWords) == 1:
            return thresholdZeroWords[0]
        elif len(thresholdZeroWords) > 1:
            return np.random.choice(thresholdZeroWords)
        
        #Otherwise, consider all categories as possibilities
        weightedActs = np.zeros(len(candidates))
        for i in range(len(candidates)):         
            
            #Get the word for each category
            word = candidates[i]
            
            #Compute the activation of each category
            weightedAct = (self.getLandscapeValue(word.category.actSpace, exemplar.value))
            
            #For real words, perform weighting using given recognition threshold
            if word.frequency != 0:
                recogThreshold = self.recogThreshold[word.frequency]
                
                #Threshold 1 means no weighting
                if recogThreshold != 1:
                    weightedAct /= recogThreshold
                    
            #For nonwords, multiply the activation by the nonword competition factor
            else:
                
                #Threshold 1 means no weighting
                if self.nonwordCompetition != 1:
                    weightedAct *= self.nonwordCompetition
                
            weightedActs[i] = weightedAct
        
        #Convert weighted activations to probability distribution
        probabilities = weightedActs / np.sum(weightedActs)
        probabilities[probabilities < 0] = 0
        probabilities[probabilities > 1] = 1
        
        #Choose a word to recognize based on the probability distribution
        recognizedWord = np.random.choice(candidates, p=probabilities)
        
        return recognizedWord
        
        
    def typicalityEvaluation(self, exemplar, word):
        """Assess whether an exemplar passes the typicality evaluation as an
        instance of a given word. Nonwords never pass.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object.
        word: the WordType object that the exemplar has been recognized as.
        =========
        Return: a Boolean indicating whether the exemplar passes the evaluation.
        """

        #Remove nonwords
        if word.frequency == 0:
            return False

        #Set storage threshold based on word frequency and convert to an absolute activation value
        storeThreshold = self.storeThreshold[word.frequency] * word.category.totalWordFrequency
        
        #Pass automatically if activation threshold is zero
        if storeThreshold == 0:
            return True
        
        #Otherwise make a random choice, using storage threshold
        else:
            r = np.random.uniform()
            return r < (1 - np.exp(-np.log(2) * self.getLandscapeValue(word.category.actSpace, exemplar.value) / storeThreshold))
            

    def storeExemplar(self, wordType, newExemplar):
        """Store a new exemplar of a given word type, by overwriting a random 
        exemplar of the same type.
        =========
        Arguments
        ---------
        wordType: the WordType object that the exemplar is being stored in.
        newExemplar: the Exemplar object to be stored.
        """
        
        #Choose an old exemplar to overwrite
        oldExemplar = wordType.chooseExemplar()
        
        #Overwrite the old exemplar with the new one
        wordType.swapExemplars(oldExemplar, newExemplar)
 

class CategoryCloud:
    """A representation of a category, which is a cloud of exemplars of word types
    which contain the phoneme in question.
    Contains methods for interacting with categories and their word types.
    """    
    
    def __init__(self, name, bias, agent):
        """Initialize the key attributes of the category.
        =========
        Arguments
        ---------
        name: a unique string name for the category (shared across agents).
        bias: the bias multiplier for the category.
        agent: the Agent object in which the category exists.        
        """
        
        #Backpointers
        self.agent = agent
        self.model = self.agent.model
        
        #Landscapes
        self.actSpace = np.zeros(self.model.spacePoints)
        self.proSpace = np.zeros(self.model.spacePoints)
        self.convProSpace = np.zeros(self.model.spacePoints)
        
        #Word structures
        self.wordTypeStore = dict()
        self.wordTypes = list()
        self.wordFrequencies = list()
        self.totalWordFrequency = 0.0
        
        #Apply bias multiplier to overall bias
        self.bias = bias
        if bias == "None":
            self.scaledBias = None
        else:
            self.scaledBias = bias * self.agent.biasSF
        
        #Properties
        self.name = name
        
    
    def addWordType(self, wordType):
        """Add a new word type to the category.
        =========
        Arguments
        ---------
        wordType: a WordType object.
        """
        
        #Add to word type structures
        self.wordTypeStore[wordType.frame] = wordType
        self.wordTypes.append(wordType)
        self.wordFrequencies.append(wordType.frequency)
        
        #Update total frequency counter
        self.totalWordFrequency += wordType.frequency
        
    
    def setWordProbabilities(self):
        """Change the word frequencies into an array of probabilities."""
               
        self.wordTypes = np.array(self.wordTypes)
        self.wordFrequencies = np.array(self.wordFrequencies) / self.totalWordFrequency
        
        
    def chooseWord(self):
        """Return a WordType chosen from the category according to frequency.
        =========
        Return: a WordType object.
        """
        
        wordType = np.random.choice(self.wordTypes, p=self.wordFrequencies)    
        
        return wordType
        
    
    def getWordType(self, frame):
        """Return the WordType object in the current category for the given frame.
        Return an empty WordType (freq 0) if there is no word with that frame in 
        this category.
        =========
        Arguments
        ---------
        frame: a string representation of the phonological frame.
        =========
        Return: a WordType object.
        """
        
        #Look up word type if it exists
        if frame in self.wordTypeStore:        
            return self.wordTypeStore[frame]
            
        #Otherwise, return empty word type
        return WordType(frame, 0, self)
        
        
    def updateActivation(self, centerPoint, operation):
        """Adds or subtracts activation centered at the given point in the exemplar space.
        =========
        Arguments
        ---------
        centerPoint: the point in the acoustic space at which to center the addition/
                     subtraction of activation.
        operation: a pointwise NumPy operation, np.add or np.subtract.   
        """
        
        #Add activation
        self.actSpace = operation(self.actSpace, self.agent.perWindows[centerPoint])
        
        #Also update production landscapes if there is entrenchment
        if self.agent.proWindowSD != 0:
            self.proSpace = operation(self.proSpace, self.agent.proWindows[centerPoint])
            self.convProSpace = operation(self.convProSpace, self.agent.convProWindows[centerPoint])  
        
        
    def shuffleWords(self):
        """Shuffle the exemplars across words within a frequency bin."""
        
        #Get filters for each frequency class
        freqFilters = {"HF": self.model.tools.hfFilter,
                       "MHF": self.model.tools.mhfFilter,
                       "MLF": self.model.tools.mlfFilter,
                       "LF": self.model.tools.lfFilter}
                       
        #Create a set of words in each frequency class
        freqBinWords = {freq: set() for freq in freqFilters}
        for word in self.wordTypes:
            for freq in ("MLF", "MHF", "LF", "HF"): #This order is optimized for fast lookup
                if freqFilters[freq](word):
                    freqBinWords[freq].add(word)
                    break
        
        #Create a set of exemplars in each frequency class
        freqBinExemplars = {freq: list() for freq in freqBinWords}
        for freq in freqBinWords:
            for word in freqBinWords[freq]:
                freqBinExemplars[freq] += word.exemplars
                
            #Shuffle the exemplars within the class
            random.shuffle(freqBinExemplars[freq])
            
            #Reassign exemplars to words from the class
            for word in freqBinWords[freq]:
                word.createNewExemplarStore()
                for exemplarNumber in range(word.frequency):
                    exemplar = freqBinExemplars[freq].pop()
                    word.addExemplarWithoutUpdating(exemplar)
                    
    
    def shuffleFrequencies(self):
        """Shuffle frequencies across words within the category."""
        
        #Create paired list of exemplar stores and word frequencies
        exemplarStoreAndFreqList = [(word.exemplars, word.exemplarsToPosition, word.frequency) for word in self.wordTypes]
        
        #Shuffle a list of indices mapping words to exemplars / frequencies
        wordIndices = np.arange(len(self.wordTypes))
        random.shuffle(wordIndices)
        
        #Re-map words to frequencies and exemplars
        self.wordFrequencies = list()
        for i in range(len(self.wordTypes)):
            word = self.wordTypes[i]
            (word.exemplars, word.exemplarsToPosition, word.frequency) = exemplarStoreAndFreqList[wordIndices[i]]
            self.wordFrequencies.append(word.frequency)
            
        #Reset word probabilities
        self.setWordProbabilities()
                    
                    
    def clearExemplars(self):
        """Clear all of the exemplars in the category."""
        
        #Clear activations
        self.actSpace = np.zeros(self.model.spacePoints)
        self.proSpace = np.zeros(self.model.spacePoints)
        self.convProSpace = np.zeros(self.model.spacePoints)
        
        #Remove exemplars
        for word in self.wordTypes:
            word.createNewExemplarStore()


class WordType:
    """A representation of a word type, which is a cloud of exemplars.
    Contains methods for interacting with word types and their exemplars.
    """     
    
    def __init__(self, frame, frequency, category):
        """Initialize the key attributes of the word type.
        =========
        Arguments
        ---------
        frame: a string representing the phonological frame of the word,
               which is unique within the category.
        frequency: the (log) frequency of the word type.
        category: the CategoryCloud object to which the word type belongs.        
        """        
        
        #Backpointers
        self.category = category 
        self.agent = self.category.agent
        self.model = self.agent.model       
        
        #Properties
        self.frequency = frequency
        self.frame = frame
        
        #Exemplar structure
        self.createNewExemplarStore()
    
    
    def createNewExemplarStore(self):
        """Create a store for exemplars. Exemplars are stored in a list with 
        indices in a linked dict, for efficient addition, removal, and sampling."""
        
        self.exemplars = list()
        self.exemplarsToPosition = dict()
    
    
    def addExemplar(self, exemplar):
        """Add an exemplar to the word store, and update category activation.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be added.
        """
        
        self.addExemplarWithoutUpdating(exemplar)
        self.category.updateActivation(exemplar.value, np.add)
        
        
    def removeExemplar(self, exemplar):
        """Remove a single exemplar and update category activation.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be removed.
        """
        
        self.removeExemplarWithoutUpdating(exemplar)            
        self.category.updateActivation(exemplar.value, np.subtract)
        
        
    def addExemplarWithoutUpdating(self, exemplar):
        """Add an exemplar to the word store, without updating category activation.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be added.
        """
        
        #Fill in the backpointers for the exemplar
        exemplar.word = self
        exemplar.category = self.category
        exemplar.agent = self.agent
        exemplar.model = self.model
        
        #Add to word type store
        self.exemplars.append(exemplar)
        self.exemplarsToPosition[exemplar] = len(self.exemplars)-1
        
        
    def removeExemplarWithoutUpdating(self, exemplar):
        """Remove a single exemplar without updating category activation.
        =========
        Arguments
        ---------
        exemplar: the Exemplar object to be removed.
        """
        
        #Remove the last exemplar and swap it with the exemplar to be removed
        position = self.exemplarsToPosition.pop(exemplar)
        lastExemplar = self.exemplars.pop()
        if position != len(self.exemplars):
            self.exemplars[position] = lastExemplar
            self.exemplarsToPosition[lastExemplar] = position 
            
            
    def swapExemplars(self, oldExemplar, newExemplar):
        """Swap an old exemplar in the store for a new exemplar, and update
        category activation accordingly.
        =========
        Arguments
        ---------
        oldExemplar: the Exemplar object to be removed.
        newExemplar: the Exemplar object to be added in place.
        """
        
        #Swap exemplars
        position = self.exemplarsToPosition.pop(oldExemplar)
        self.exemplars[position] = newExemplar
        self.exemplarsToPosition[newExemplar] = position
        
        #Update activations
        self.category.updateActivation(oldExemplar.value, np.subtract)
        self.category.updateActivation(newExemplar.value, np.add)
        
              
    def chooseExemplar(self):
        """Randomly choose an exemplar from the store.
        =========
        Return: an Exemplar object.
        """
        
        #Exemplar choice follows a uniform distribution
        exemplar = np.random.choice(self.exemplars)
            
        return exemplar      
        
        

class Exemplar:
    """A representation of an exemplar, which is an acoustic value within a phonological frame.
    Contains methods for interacting with exemplars.
    """  
    
    def __init__(self, value, frame, speaker):
        """Create a new exemplar.
        =========
        Arguments
        ---------
        value: the acoustic value of the exemplar.
        frame: a string representation of the phonological frame of the exemplar.
        speaker: the Agent object who uttered the exemplar.
        """
        
        #Properties
        self.value = value           
        self.wordFrame = frame 
        self.speaker = speaker
        
        #Backpointers
        self.word = None
        self.category = None
        self.agent = None
        self.model = None
        if self.speaker is not None:
            self.model = self.speaker.model         
       
         
    def discretize(self):
        """Discretize the value of a given exemplar."""
        
        #Discretization is rounding to 2 decimal places.
        self.value = round(np.round(float(self.value) / self.model.spaceStep) * self.model.spaceStep, 2)
        
        
    def bringIntoSpace(self):
        """Bring the value of an exemplar back into the space of the model, if it has left."""
        
        self.value = min(max(self.value, self.model.spaceMin), self.model.spaceMax)